
#include <stdio.h>
#include <stdlib.h>

#ifdef FREETYPE_HDR_DIRECTORY
#include <freetype/freetype.h>
#else
#include <freetype.h>
#endif

#include "SDL.h"
#include "SDL_ttf.h"

/* Macro to convert a character to a Unicode value -- assume already Unicode */
#define UNICODE(c)	c

/* The structure used to hold internal font information */
struct _TTF_Font {
	TT_Face face;
	TT_Instance inst;
	TT_Glyph glyph;
	TT_CharMap map;
	TT_Raster_Map scratch;
};

/* The FreeType font engine */
static TT_Engine engine;

int TTF_Init(void)
{
	int error;

	error = TT_Init_FreeType(&engine);
	if ( error ) {
		SDL_SetError("Couldn't init FreeType engine");
		return(-1);
	}
	return(0);
}

TTF_Font *TTF_OpenFont(const char *file, int ptsize)
{
	TTF_Font *font;
	TT_Instance_Metrics imetrics;
	int i, n;
	TT_UShort platform, encoding;
	int error;

	font = (TTF_Font *)malloc(sizeof(*font));
	if ( font == NULL ) {
		SDL_SetError("Out of memory");
		return(NULL);
	}
	memset(font, 0, sizeof(*font));

	/* Open the font and create ancillary data */
	error = TT_Open_Face(engine, file, &font->face);
	if ( error ) {
		SDL_SetError("Couldn't load font file");
		free(font);
		return(NULL);
	}
	error = TT_New_Glyph(font->face, &font->glyph);
	if ( error ) {
		SDL_SetError("Couldn't create glyph container");
		TTF_CloseFont(font);
		return(NULL);
	}
	error = TT_New_Instance(font->face, &font->inst);
	if ( error ) {
		SDL_SetError("Couldn't create font instance");
		TTF_CloseFont(font);
		return(NULL);
	}

	/* Set the display resolution */
	error = TT_Set_Instance_Resolutions(font->inst, 72, 72);
	if ( error ) {
		SDL_SetError("Couldn't set font resolution");
		TTF_CloseFont(font);
		return(NULL);
	}
	error = TT_Set_Instance_CharSize(font->inst, ptsize*64);
	if ( error ) {
		SDL_SetError("Couldn't set font size");
		TTF_CloseFont(font);
		return(NULL);
	}

	/* Get a Unicode mapping for this font */
	n = TT_Get_CharMap_Count(font->face);
	for ( i=0; i<n; ++i ) {
		TT_Get_CharMap_ID(font->face, i, &platform, &encoding);
		if ( ((platform == TT_PLATFORM_MICROSOFT) &&
		                  (encoding == TT_MS_ID_UNICODE_CS)) ||
		     ((platform == TT_PLATFORM_APPLE_UNICODE) &&
		                  (encoding == TT_APPLE_ID_DEFAULT)) ) {
			TT_Get_CharMap(font->face, i, &font->map);
			break;
		}
	}
	if ( i == n ) {
		SDL_SetError("Font doesn't have a Unicode mapping");
		TTF_CloseFont(font);
		return(NULL);
	}

	/* Create a scratch pixmap for this font */
	TT_Get_Instance_Metrics(font->inst, &imetrics);
	font->scratch.rows = imetrics.y_ppem + 32;
	font->scratch.width = imetrics.x_ppem + 32;
	font->scratch.flow = TT_Flow_Up;
	font->scratch.cols = (font->scratch.width+3) & -4;
	font->scratch.size = font->scratch.rows*font->scratch.cols;
	font->scratch.bitmap = malloc((int)font->scratch.size);
	if ( font->scratch.bitmap == NULL ) {
		SDL_SetError("Out of memory");
		TTF_CloseFont(font);
		return(NULL);
	}
	return(font);
}

void TTF_CloseFont(TTF_Font *font)
{
	TT_Close_Face(font->face);
	if ( font->scratch.bitmap ) {
		free(font->scratch.bitmap);
	}
	free(font);
}

static Uint16 *ASCII_to_UNICODE(Uint16 *unicode, const char *text, int len)
{
	int i;

	for ( i=0; i < len; ++i ) {
		unicode[i] = ((const unsigned char *)text)[i];
	}
	unicode[i] = 0;

	return unicode;
}

static Uint16 *UTF8_to_UNICODE(Uint16 *unicode, const char *utf8, int len)
{
	int i, j;
	Uint16 ch;

	for ( i=0, j=0; i < len; ++i, ++j ) {
		ch = ((const unsigned char *)utf8)[i];
		if ( ch >= 0xF0 ) {
			ch  =  (Uint16)(utf8[i]&0x07) << 18;
			ch |=  (Uint16)(utf8[++i]&0x3F) << 12;
			ch |=  (Uint16)(utf8[++i]&0x3F) << 6;
			ch |=  (Uint16)(utf8[++i]&0x3F);
		} else
		if ( ch >= 0xE0 ) {
			ch  =  (Uint16)(utf8[i]&0x3F) << 12;
			ch |=  (Uint16)(utf8[++i]&0x3F) << 6;
			ch |=  (Uint16)(utf8[++i]&0x3F);
		} else
		if ( ch >= 0xC0 ) {
			ch  =  (Uint16)(utf8[i]&0x3F) << 6;
			ch |=  (Uint16)(utf8[++i]&0x3F);
		}
		unicode[j] = ch;
	}
	unicode[j] = 0;

	return unicode;
}

int TTF_FontHeight(TTF_Font *font)
{
	return(font->scratch.rows - 32);
}

int TTF_SizeText(TTF_Font *font, const char *text,
			int *w, int *h, int *descent)
{
	Uint16 *unicode_text;
	int unicode_len;
	int status;

	/* Copy the Latin-1 text to a UNICODE text buffer */
	unicode_len = strlen(text);
	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
	if ( unicode_text == NULL ) {
		SDL_SetError("Out of memory");
		return -1;
	}
	ASCII_to_UNICODE(unicode_text, text, unicode_len);

	/* Render the new text */
	status = TTF_SizeUNICODE(font, unicode_text, w, h, descent);

	/* Free the text buffer and return */
	free(unicode_text);
	return status;
}

int TTF_SizeUTF8(TTF_Font *font, const char *text,
			int *w, int *h, int *descent)
{
	Uint16 *unicode_text;
	int unicode_len;
	int status;

	/* Copy the UTF-8 text to a UNICODE text buffer */
	unicode_len = strlen(text);
	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
	if ( unicode_text == NULL ) {
		SDL_SetError("Out of memory");
		return -1;
	}
	UTF8_to_UNICODE(unicode_text, text, unicode_len);

	/* Render the new text */
	status = TTF_SizeUNICODE(font, unicode_text, w, h, descent);

	/* Free the text buffer and return */
	free(unicode_text);
	return status;
}

int TTF_SizeUNICODE(TTF_Font *font, const Uint16 *text,
			int *w, int *h, int *descent)
{
	int status;
	const Uint16 *ch;
	int index;
	TT_Glyph_Metrics metrics;
	TT_F26Dot6 xoff, z, minx, miny, maxx, maxy;
	TT_Error error;

	/* Initialize everything to 0 */
	status = 0;
	minx = miny = maxx = maxy = 0;

	/* Load each character and sum it's bounding box */
	xoff = 0;
	for ( ch=text; *ch; ++ch ) {
		index = TT_Char_Index(font->map, *ch);
		if ( ! (error=TT_Load_Glyph(font->inst, font->glyph,
						index, TTLOAD_DEFAULT)) ) {
			TT_Get_Glyph_Metrics(font->glyph, &metrics);

			z = xoff + metrics.bbox.xMin;
			if ( minx > z )
				minx = z;

			z = xoff + metrics.bbox.xMax;
			if ( maxx < z )
				maxx = z;

			z = metrics.bbox.yMin;
			if ( miny > z )
				miny = z;

			z = metrics.bbox.yMax;
			if ( maxy < z )
				maxy = z;

			xoff += metrics.advance & -64;
		}
	}

	/* Fill the bounds rectangle */
	if ( w ) {
		*w = (maxx-minx)/64;
	}
	if ( h ) {
		*h = (maxy-miny)/64;
	}
	if ( descent ) {
		*descent = -miny/64;
	}
	return status;
}

/* Convert the Latin-1 text to UNICODE and render it
*/
SDL_Surface *TTF_RenderText_Solid(TTF_Font *font,
				const char *text, SDL_Color fg)
{
	SDL_Surface *textbuf;
	Uint16 *unicode_text;
	int unicode_len;

	/* Copy the Latin-1 text to a UNICODE text buffer */
	unicode_len = strlen(text);
	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
	if ( unicode_text == NULL ) {
		SDL_SetError("Out of memory");
		return(NULL);
	}
	ASCII_to_UNICODE(unicode_text, text, unicode_len);

	/* Render the new text */
	textbuf = TTF_RenderUNICODE_Solid(font, unicode_text, fg);

	/* Free the text buffer and return */
	free(unicode_text);
	return(textbuf);
}

/* Convert the UTF-8 text to UNICODE and render it
*/
SDL_Surface *TTF_RenderUTF8_Solid(TTF_Font *font,
				const char *text, SDL_Color fg)
{
	SDL_Surface *textbuf;
	Uint16 *unicode_text;
	int unicode_len;

	/* Copy the UTF-8 text to a UNICODE text buffer */
	unicode_len = strlen(text);
	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
	if ( unicode_text == NULL ) {
		SDL_SetError("Out of memory");
		return(NULL);
	}
	UTF8_to_UNICODE(unicode_text, text, unicode_len);

	/* Render the new text */
	textbuf = TTF_RenderUNICODE_Solid(font, unicode_text, fg);

	/* Free the text buffer and return */
	free(unicode_text);
	return(textbuf);
}

SDL_Surface *TTF_RenderUNICODE_Solid(TTF_Font *font,
				const Uint16 *text, SDL_Color fg)
{
	SDL_Surface *textbuf;
	const Uint16 *ch;
	int index;
	TT_Glyph_Metrics metrics;
	TT_F26Dot6 xoff, z, minx, miny, maxx, maxy;
	int descent;
	SDL_Palette *palette;
	int rdiff, gdiff, bdiff;
	int x, w, h;
	Uint8 *srcbuf;
	Uint16 srcpitch;
	Uint8 *dstbuf;
	Uint16 dstpitch;
	TT_Error error;

	/* Fill the bounds rectangle */
	TTF_SizeUNICODE(font, text, &w, &h, &descent);
#ifdef TEXT_DEBUG
fprintf(stderr, "Text block: %dx%d with %d descent\n", w, h, descent);
#endif
	/* Allocate an 8-bit pixmap */
	w = (w+7)&~7;
	textbuf = SDL_AllocSurface(SDL_SWSURFACE, w, h, 8, 0, 0, 0, 0);
	if ( textbuf == NULL ) {
		return(NULL);
	}

	/* Fill the palette with the foreground color */
	palette = textbuf->format->palette;
	palette->colors[0].r = 255-fg.r;
	palette->colors[0].g = 255-fg.g;
	palette->colors[0].b = 255-fg.b;
	palette->colors[1].r = fg.r;
	palette->colors[1].g = fg.g;
	palette->colors[1].b = fg.b;
	SDL_SetColorKey(textbuf, SDL_SRCCOLORKEY, 0);

	/* Load and render each character */
	xoff = 0;
	for ( ch=text; *ch; ++ch ) {
		index = TT_Char_Index(font->map, UNICODE(*ch));
		if ( ! (error=TT_Load_Glyph(font->inst, font->glyph,
						index, TTLOAD_DEFAULT)) ) {

			/* Get the bounding box and glyph pixmap */
			TT_Get_Glyph_Metrics(font->glyph, &metrics);
			minx = metrics.bbox.xMin;
			maxx = metrics.bbox.xMax;
			miny = metrics.bbox.yMin;
			maxy = metrics.bbox.yMax;
#ifdef TEXT_DEBUG
fprintf(stderr, "Rendering '%c'  %dx%d minx = %d, miny = %d at x = %d\n",
	*ch, (maxx-minx)/64, (maxy-miny)/64, minx/64, miny/64, xoff/64);
#endif
			memset(font->scratch.bitmap, 0, font->scratch.size);
			TT_Get_Glyph_Bitmap(font->glyph,
						&font->scratch, -minx, -miny);

			/* OR the pixmap into the text surface */
			w = (maxx-minx)/64;
			if ( w > font->scratch.cols )
				w = font->scratch.cols;
			if ( (xoff == 0) && minx < 0 )
				minx = 0;
			h = (maxy-miny)/64;
			if ( h > font->scratch.rows )
				h = font->scratch.rows;
			srcbuf = font->scratch.bitmap;
			srcpitch = font->scratch.cols - (w+7)/8;
			dstbuf = (Uint8 *)textbuf->pixels +
				(textbuf->h-descent-miny/64-1)*textbuf->pitch +
								(xoff+minx)/64;
			dstpitch = textbuf->pitch + (w+7)&~7;
			while ( h-- ) {
				for ( x = 0; x < w; x += 8 ) {
					Uint8 c = *srcbuf++;
					*dstbuf++ |= (c&0x80)>>7;
					c <<= 1;
					*dstbuf++ |= (c&0x80)>>7;
					c <<= 1;
					*dstbuf++ |= (c&0x80)>>7;
					c <<= 1;
					*dstbuf++ |= (c&0x80)>>7;
					c <<= 1;
					*dstbuf++ |= (c&0x80)>>7;
					c <<= 1;
					*dstbuf++ |= (c&0x80)>>7;
					c <<= 1;
					*dstbuf++ |= (c&0x80)>>7;
					c <<= 1;
					*dstbuf++ |= (c&0x80)>>7;
				}
				srcbuf += srcpitch;
				dstbuf -= dstpitch;
			}
			xoff += metrics.advance & -64;
		}
	}
	return(textbuf);
}

/* Convert the Latin-1 text to UNICODE and render it
*/
SDL_Surface *TTF_RenderText_Shaded(TTF_Font *font,
				const char *text, SDL_Color fg, SDL_Color bg)
{
	SDL_Surface *textbuf;
	Uint16 *unicode_text;
	int unicode_len;

	/* Copy the Latin-1 text to a UNICODE text buffer */
	unicode_len = strlen(text);
	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
	if ( unicode_text == NULL ) {
		SDL_SetError("Out of memory");
		return(NULL);
	}
	ASCII_to_UNICODE(unicode_text, text, unicode_len);

	/* Render the new text */
	textbuf = TTF_RenderUNICODE_Shaded(font, unicode_text, fg, bg);

	/* Free the text buffer and return */
	free(unicode_text);
	return(textbuf);
}

/* Convert the UTF-8 text to UNICODE and render it
*/
SDL_Surface *TTF_RenderUTF8_Shaded(TTF_Font *font,
				const char *text, SDL_Color fg, SDL_Color bg)
{
	SDL_Surface *textbuf;
	Uint16 *unicode_text;
	int unicode_len;

	/* Copy the UTF-8 text to a UNICODE text buffer */
	unicode_len = strlen(text);
	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
	if ( unicode_text == NULL ) {
		SDL_SetError("Out of memory");
		return(NULL);
	}
	UTF8_to_UNICODE(unicode_text, text, unicode_len);

	/* Render the new text */
	textbuf = TTF_RenderUNICODE_Shaded(font, unicode_text, fg, bg);

	/* Free the text buffer and return */
	free(unicode_text);
	return(textbuf);
}

SDL_Surface *TTF_RenderUNICODE_Shaded(TTF_Font *font,
				const Uint16 *text, SDL_Color fg, SDL_Color bg)
{
	SDL_Surface *textbuf;
	const Uint16 *ch;
	int index;
	TT_Glyph_Metrics metrics;
	TT_F26Dot6 xoff, z, minx, miny, maxx, maxy;
	int descent;
	SDL_Palette *palette;
	int rdiff, gdiff, bdiff;
	int x, w, h;
	Uint8 *srcbuf;
	Uint16 srcpitch;
	Uint8 *dstbuf;
	Uint16 dstpitch;
	TT_Error error;

	/* Fill the bounds rectangle */
	TTF_SizeUNICODE(font, text, &w, &h, &descent);
#ifdef TEXT_DEBUG
fprintf(stderr, "Text block: %dx%d with %d descent\n", w, h, descent);
#endif
	/* Allocate an 8-bit pixmap */
	textbuf = SDL_AllocSurface(SDL_SWSURFACE, w, h, 8, 0, 0, 0, 0);
	if ( textbuf == NULL ) {
		return(NULL);
	}

	/* Fill the palette with 5 levels of shading from bg to fg */
	palette = textbuf->format->palette;
	rdiff = fg.r - bg.r;
	gdiff = fg.g - bg.g;
	bdiff = fg.b - bg.b;
	for ( index=0; index<5; ++index ) {
		palette->colors[index].r = bg.r + (index*rdiff)/4;
		palette->colors[index].g = bg.g + (index*gdiff)/4;
		palette->colors[index].b = bg.b + (index*bdiff)/4;
	}
	/* The other 3 levels are used as overflow when ORing pixels */
	for ( ; index<8; ++index ) {
		palette->colors[index] = palette->colors[4];
	}

	/* Load and render each character */
	xoff = 0;
	for ( ch=text; *ch; ++ch ) {
		index = TT_Char_Index(font->map, UNICODE(*ch));
		if ( ! (error=TT_Load_Glyph(font->inst, font->glyph,
						index, TTLOAD_DEFAULT)) ) {

			/* Get the bounding box and glyph pixmap */
			TT_Get_Glyph_Metrics(font->glyph, &metrics);
			minx = metrics.bbox.xMin;
			maxx = metrics.bbox.xMax;
			miny = metrics.bbox.yMin;
			maxy = metrics.bbox.yMax;
#ifdef TEXT_DEBUG
fprintf(stderr, "Rendering '%c'  %dx%d minx = %d, miny = %d at x = %d\n",
	*ch, (maxx-minx)/64, (maxy-miny)/64, minx/64, miny/64, xoff/64);
#endif
			memset(font->scratch.bitmap, 0, font->scratch.size);
			TT_Get_Glyph_Pixmap(font->glyph,
						&font->scratch, -minx, -miny);

			/* OR the pixmap into the text surface */
			w = (maxx-minx)/64;
			if ( w > font->scratch.cols )
				w = font->scratch.cols;
			if ( (xoff == 0) && minx < 0 )
				minx = 0;
			h = (maxy-miny)/64;
			if ( h > font->scratch.rows )
				h = font->scratch.rows;
			srcbuf = font->scratch.bitmap;
			srcpitch = font->scratch.cols - w;
			dstbuf = (Uint8 *)textbuf->pixels +
				(textbuf->h-descent-miny/64-1)*textbuf->pitch +
								(xoff+minx)/64;
			dstpitch = textbuf->pitch + w;
			while ( h-- ) {
				for ( x = 0; x < w; ++x )
					*dstbuf++ |= *srcbuf++;
				srcbuf += srcpitch;
				dstbuf -= dstpitch;
			}
			xoff += metrics.advance & -64;
		}
	}
	return(textbuf);
}

/* Convert the Latin-1 text to UNICODE and render it
*/
SDL_Surface *TTF_RenderText_Blended(TTF_Font *font,
				const char *text, SDL_Color fg)
{
	SDL_Surface *textbuf;
	Uint16 *unicode_text;
	int unicode_len;

	/* Copy the Latin-1 text to a UNICODE text buffer */
	unicode_len = strlen(text);
	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
	if ( unicode_text == NULL ) {
		SDL_SetError("Out of memory");
		return(NULL);
	}
	ASCII_to_UNICODE(unicode_text, text, unicode_len);

	/* Render the new text */
	textbuf = TTF_RenderUNICODE_Blended(font, unicode_text, fg);

	/* Free the text buffer and return */
	free(unicode_text);
	return(textbuf);
}

/* Convert the UTF-8 text to UNICODE and render it
*/
SDL_Surface *TTF_RenderUTF8_Blended(TTF_Font *font,
				const char *text, SDL_Color fg)
{
	SDL_Surface *textbuf;
	Uint16 *unicode_text;
	int unicode_len;

	/* Copy the UTF-8 text to a UNICODE text buffer */
	unicode_len = strlen(text);
	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
	if ( unicode_text == NULL ) {
		SDL_SetError("Out of memory");
		return(NULL);
	}
	UTF8_to_UNICODE(unicode_text, text, unicode_len);

	/* Render the new text */
	textbuf = TTF_RenderUNICODE_Blended(font, unicode_text, fg);

	/* Free the text buffer and return */
	free(unicode_text);
	return(textbuf);
}

SDL_Surface *TTF_RenderUNICODE_Blended(TTF_Font *font,
				const Uint16 *text, SDL_Color fg)
{
	SDL_Surface *textbuf;
	const Uint16 *ch;
	int index;
	TT_Glyph_Metrics metrics;
	TT_F26Dot6 xoff, z, minx, miny, maxx, maxy;
	int descent;
	int alpha;
	int x, w, h;
	Uint8 *srcbuf;
	int srcpitch;
	Uint32 *dstbuf;
	int dstpitch;
	TT_Error error;

	/* Fill the bounds rectangle */
	TTF_SizeUNICODE(font, text, &w, &h, &descent);
#ifdef TEXT_DEBUG
fprintf(stderr, "Text block: %dx%d with %d descent\n", w, h, descent);
#endif
	/* Allocate a 32-bit ARGB pixmap */
	textbuf = SDL_AllocSurface(SDL_SWSURFACE, w, h, 32,
	                           0x00FF0000, 0x0000FF00, 0x000000FF,
	                           0xFF000000);
	if ( textbuf == NULL ) {
		return(NULL);
	}
	/* Set surface completely transparent */
	memset(textbuf->pixels, 0xFF, (textbuf->h*textbuf->pitch));

	/* Load and render each character */
	xoff = 0;
	for ( ch=text; *ch; ++ch ) {
		index = TT_Char_Index(font->map, UNICODE(*ch));
		if ( ! (error=TT_Load_Glyph(font->inst, font->glyph,
						index, TTLOAD_DEFAULT)) ) {

			/* Get the bounding box and glyph pixmap */
			TT_Get_Glyph_Metrics(font->glyph, &metrics);
			minx = metrics.bbox.xMin;
			maxx = metrics.bbox.xMax;
			miny = metrics.bbox.yMin;
			maxy = metrics.bbox.yMax;
#ifdef TEXT_DEBUG
fprintf(stderr, "Rendering '%c'  %dx%d minx = %d, miny = %d at x = %d\n",
	*ch, (maxx-minx)/64, (maxy-miny)/64, minx/64, miny/64, xoff/64);
#endif
			memset(font->scratch.bitmap, 0, font->scratch.size);
			TT_Get_Glyph_Pixmap(font->glyph,
						&font->scratch, -minx, -miny);

			/* OR the pixmap into the text surface */
			w = (maxx-minx)/64;
			if ( w > font->scratch.cols )
				w = font->scratch.cols;
			if ( (xoff == 0) && minx < 0 )
				minx = 0;
			h = (maxy-miny)/64;
			if ( h > font->scratch.rows )
				h = font->scratch.rows;
			srcbuf = font->scratch.bitmap;
			srcpitch = font->scratch.cols - w;
			dstbuf = (Uint32 *)textbuf->pixels +
			      (textbuf->h-descent-miny/64-1)*textbuf->pitch/4 +
								(xoff+minx)/64;
			dstpitch = textbuf->pitch/4 + w;
			while ( h-- ) {
				for ( x = 0; x < w; ++x ) {
					if ( *srcbuf ) {
						alpha = *dstbuf>>24;
						alpha -= (256*(*srcbuf))/4;
						if ( alpha < 0 ) {
							alpha = 0;
						}
						*dstbuf = (alpha<<24)|
						     (fg.r<<16)|(fg.g<<8)|fg.b;
					}
					++srcbuf;
					++dstbuf;
				}
				srcbuf += srcpitch;
				dstbuf -= dstpitch;
			}
			xoff += metrics.advance & -64;
		}
	}
	return(textbuf);
}

void TTF_Quit(void)
{
	TT_Done_FreeType(engine);
}
